// Copyright 2023 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package jobqueue

import (
	"context"
	"errors"
	"gorm.io/gorm"
	"math/rand"
	"sync"
	"time"
)

type DBSaverStatus int

const (
	DBSaverStatusStopped = iota
	DBSaverStatusRunning
)

type DBSaver struct {
	saveChannel     chan GenericJob
	stopChan        chan struct{}
	migrateFlag     bool
	manager         *Manager
	status          DBSaverStatus
	mu              sync.Mutex
	jobSaveProgress sync.WaitGroup
}

type RunnerData string
type SchedulerData string

// NewDBSaver creates a new DBSaver
func NewDBSaver() *DBSaver {
	return &DBSaver{
		saveChannel: make(chan GenericJob, 1000),
		stopChan:    make(chan struct{}),
	}
}

// SetManager sets the manager of the DBSaver
func (s *DBSaver) SetManager(manager *Manager) *DBSaver {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.manager = manager
	return s
}

func (s *DBSaver) setStatus(status DBSaverStatus) *DBSaver {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.status = status
	return s
}

// isStatus returns true if the DBSaver has the given status
// the lock is not needed here, because it is only used in the Start() method
func (s *DBSaver) isStatus(status DBSaverStatus, lock bool) bool {
	if lock {
		s.mu.Lock()
		defer s.mu.Unlock()
	}

	return s.status == status

}

func StartDBSaver[P *Promise[bool]](s *DBSaver) *Promise[bool] {
	s.mu.Lock()
	defer s.mu.Unlock()

	return NewPromise[bool](func(resolve func(bool), reject func(error)) {

		if s.manager == nil || s.manager.database == nil {
			reject(ErrNoDatabaseConnection)
			return
		}

		if s.isStatus(DBSaverStatusRunning, false) {
			resolve(true)
			return
		}

		db := s.manager.database
		if !s.migrateFlag {
			err := db.AutoMigrate(&JobPersistence{}, &JobLog{}, &JobStats{})
			if err != nil {
				reject(err)
				return
			}
			s.migrateFlag = true
		}

		ready := make(chan struct{})
		go runSaver(s, db, ready)

		<-ready
		resolve(true)
	})
}

//
//// Start starts the DBSaver
//func (s *DBSaver) Start() error {
//	s.mu.Lock()
//	defer s.mu.Unlock()
//
//	if s.manager == nil || s.manager.database == nil {
//		return ErrNoDatabaseConnection
//	}
//
//	if s.isStatus(DBSaverStatusRunning, false) {
//		return nil
//	}
//
//	db := s.manager.database
//
//	if !s.migrateFlag {
//		err := db.AutoMigrate(&JobPersistence{}, &JobLog{}, &JobStats{})
//		if err != nil {
//			return err
//		}
//		s.migrateFlag = true
//	}
//
//	go runSaver(s, db)
//	return nil
//}

func runSaver(s *DBSaver, db *gorm.DB, ready chan struct{}) {
	s.setStatus(DBSaverStatusRunning)

	defer func() {
		// this runs after the function returns
		// and needs to be protected by the lock
		// of the setStatus method
		//s.status = DBSaverStatusStopped
		s.setStatus(DBSaverStatusStopped)
	}()

	close(ready)
	for {

		select {
		case job := <-s.saveChannel:
			s.jobSaveProgress.Add(1)

			err := CreateOrUpdateJob(job, db)

			if err != nil {
				Error("Error while saving job", "error", err)
			}

			s.jobSaveProgress.Done()

		case <-s.stopChan:
			return
		}
	}
}

func CreateOrUpdateJob(job GenericJob, db *gorm.DB) error {

	return db.Transaction(func(tx *gorm.DB) error {

		permJob := job.GetPersistence()

		memLogs := permJob.Logs
		permJob.Logs = nil

		var existingJob JobPersistence
		result := tx.Unscoped().Where("id = ?", permJob.GetID()).First(&existingJob)

		if result.Error != nil {
			if errors.Is(result.Error, gorm.ErrRecordNotFound) {
				err := tx.Create(&permJob).Error
				if err != nil {
					Trace("Error while creating job", "error", err)
					return err
				}
			} else {
				Trace("Error while creating job", "error", result.Error)
				return result.Error
			}
		} else {

			db.Unscoped().Model(&existingJob).Update("deleted_at", nil)

			tx.Model(&existingJob.Scheduler).Select(
				[]string{
					"type",
					"interval",
					"spec",
					"delay",
					"event",
					"time",
					"executed",
				}).UpdateColumns(SchedulerPersistence{
				Type:     "",
				Interval: 0,
				Spec:     "",
				Delay:    0,
				Event:    "",
				Time:     nil,
				Executed: false,
			})

			err := tx.Model(&existingJob).Updates(permJob).Error
			if err != nil {
				return err
			}
		}

		tx.Model(&permJob.Stats).
			Select(
				[]string{
					"run_count",
					"success_count",
					"error_count",
					"time_metrics_avg_run_time",
					"time_metrics_max_run_time",
					"time_metrics_min_run_time",
					"time_metrics_total_run_time",
				},
			).
			UpdateColumns(permJob.Stats)

		if tx.Error != nil {
			Error("Error while updating job stats", "error", tx.Error)
			return result.Error
		}

		for i, _ := range memLogs {
			memLogs[i].LogID = 0
			_ = tx.Create(&memLogs[i])
			// no error handling, if it fails, it fails
		}

		return nil

	})

}

// Stop stops the DBSaver
func (s *DBSaver) Stop() error {
	s.mu.Lock()
	defer s.mu.Unlock()

	ctx, cancel := context.WithTimeout(context.Background(), 30*time.Second)
	defer cancel()

	go func() {
		s.stopChan <- struct{}{}
		s.jobSaveProgress.Wait()
		cancel()
	}()

	<-ctx.Done()

	if errors.Is(ctx.Err(), context.DeadlineExceeded) {
		Error("DBSaver did not stop in time")
		return ctx.Err()
	}

	s.status = DBSaverStatusStopped
	return nil
}

func exponentialBackoff(retry int) time.Duration {
	waitTime := 100 * time.Millisecond
	for i := 0; i < retry; i++ {
		waitTime *= 2
		waitTime += time.Duration(rand.Int63n(int64(waitTime))) // #nosec G404
	}
	return waitTime
}

// SaveJob saves a job to the database
func (s *DBSaver) SaveJob(job GenericJob) error {
	s.mu.Lock()
	defer func() {
		if r := recover(); r != nil {
			Error("Error while saving job", "error", r)
		}
		s.mu.Unlock()
	}()

	if s.saveChannel == nil {
		return ErrDBSaverNotInitialized
	}
	if s.status != DBSaverStatusRunning {
		return ErrDBSaverNotRunning
	}

	maxRetries := 5

	for retries := maxRetries; retries > 0; retries-- {
		select {
		case s.saveChannel <- job:
			return nil
		default:
			Error("DBSaver channel is full, dropping safe for job with ID", "job_id", job.GetID())
			backoff := exponentialBackoff(maxRetries - retries)
			Trace("DBSaver channel is full, retrying in", "backoff", backoff)
			time.Sleep(backoff)
		}
	}

	return errors.New("failed to save job after multiple attempts")

}

func checkRunningSaver(s *DBSaver) (*gorm.DB, error) {
	if s.manager == nil {
		return nil, ErrNoManager
	}

	if s.manager.database == nil {
		return nil, ErrNoDatabaseConnection
	}

	if !s.isStatus(DBSaverStatusRunning, false) {
		return nil, ErrDBSaverNotRunning
	}

	return s.manager.database, nil

}

// DeleteJob deletes a job from the database
func (s *DBSaver) DeleteJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	var db *gorm.DB
	var err error

	if db, err = checkRunningSaver(s); err != nil {
		return err
	}

	s.jobSaveProgress.Add(1)
	return db.Transaction(func(tx *gorm.DB) error {
		defer s.jobSaveProgress.Done()
		permJob := job.GetPersistence()

		dbErr := tx.Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error
		if dbErr != nil {
			return dbErr
		}

		dbErr = tx.Where("job_id = ?", permJob.GetID()).Delete(&JobStats{}).Error
		if dbErr != nil {
			return dbErr
		}

		dbErr = tx.Delete(&permJob).Error
		if dbErr != nil {
			return dbErr
		}

		return nil
	})

}

func (s *DBSaver) ResetLogs(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	var db *gorm.DB
	var err error

	if db, err = checkRunningSaver(s); err != nil {
		return err
	}

	s.jobSaveProgress.Add(1)
	return db.Transaction(func(tx *gorm.DB) error {
		defer s.jobSaveProgress.Done()
		permJob := job.GetPersistence()

		// unscoped because we want to delete the logs finally
		dbErr := tx.Unscoped().Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error
		if dbErr != nil {
			return dbErr
		}

		return nil
	})
}

func (s *DBSaver) ResetStats(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.saveChannel == nil {
		return ErrDBSaverNotInitialized
	}

	if s.status != DBSaverStatusRunning {
		return ErrDBSaverNotRunning
	}

	defer func() {
		if r := recover(); r != nil {
			Error("Error while saving job", "error", r)
		}
	}()

	job.ResetStats()

	select {
	case s.saveChannel <- job:
	default:
		Error("DBSaver channel is full, dropping job with  ID", "job_id", job.GetID())
	}

	return nil
}
